import java.net.*;
import java.io.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.prefs.*;

public class T extends JFrame implements ActionListener, Runnable, DocumentListener {
    private static final String USER_PREF = "5KTwit.user";
    private static final String PASSWORD_PREF = "5KTwit.password";
    private String user = "";
    private String password = "";
    
    private JScrollPane scrollPane = null;
    private JPanel tweets = null;
    private JTextField status = null;
    private JButton update = null;
    private JButton prefs  = null;
    private JLabel charCount = null;
    
    private Document tweetsDoc = null;
    private HashSet<String> tweetIDs = new HashSet<String>(); // ids of tweets we've already seen
    
    private static final int MAX_CHAR = 140;
    
    private String lastTweetID = null;
    
    private volatile Thread currentThread = null;
    
    private JLabel statusBar = null;
    
    public T() {
        super("5KTwit");
        setDefaultCloseOperation(EXIT_ON_CLOSE);
        setPreferredSize(new Dimension(320,240));
        
        status = new JTextField(MAX_CHAR);
        update = new JButton("Update");
        prefs  = new JButton(new ImageIcon(getClass().getResource("/p")));
        charCount = new JLabel(String.valueOf(MAX_CHAR));
        
        JPanel top = new JPanel();
        top.setLayout(new BorderLayout());
        top.add(prefs, BorderLayout.WEST);
        top.add(status, BorderLayout.CENTER);
        top.add(update, BorderLayout.EAST);
        top.add(charCount, BorderLayout.SOUTH);
        
        update.addActionListener(this);
        status.addActionListener(this);
        prefs.addActionListener(this);
        
        status.getDocument().addDocumentListener(this);
        
        add(top, BorderLayout.NORTH);
        
        tweets = new JPanel();
        tweets.setLayout(new BoxLayout(tweets,BoxLayout.Y_AXIS));
        scrollPane=new JScrollPane(
            tweets,
            JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
            JScrollPane.HORIZONTAL_SCROLLBAR_NEVER
        );
        add(scrollPane,BorderLayout.CENTER);
        
        statusBar = new JLabel(" ");
        add(statusBar, BorderLayout.SOUTH);
        
        getRootPane().setDefaultButton(update);
    }
    
    public void actionPerformed(ActionEvent e) {
        if ( e.getSource() == prefs ) {
            start();
        }
        else {
            String status = this.status.getText();
            if ( status.length() <= MAX_CHAR ) {
                tweetsDoc = loadDocument("http://twitter.com/statuses/update.xml", status);
                loadTweets();
                this.status.setText("");
            }
        }
    }
    
    public void insertUpdate(DocumentEvent e) {
        updateCount();
    }
    
    public void removeUpdate(DocumentEvent e) {
        updateCount();
    }
    
    public void changedUpdate(DocumentEvent e) {
        updateCount();
    }
    
    private void updateCount() {
        int remaining = MAX_CHAR - status.getText().length();
        charCount.setText(String.valueOf(remaining));
        update.setEnabled(remaining>=0);
    }
    
    private void start() {
        getUsernameAndPassword();
        lastTweetID = null;
        tweetIDs.clear();
        tweets.removeAll();
        status.requestFocus();
        
        (currentThread = new Thread(this)).start();
    }
    
    public void run() {
        if ( SwingUtilities.isEventDispatchThread() ) {
            // update UI as running on the right thread
            loadTweets();
        }
        else {
            while( currentThread == Thread.currentThread() ) {
                String friendsTimelineURL = "http://twitter.com/statuses/friends_timeline.xml?count=200";
                if ( lastTweetID != null ) {
                    friendsTimelineURL += "&since_id=" + lastTweetID;
                }
                tweetsDoc = loadDocument(friendsTimelineURL, null);
                try {
                    // invoke and wait, to run loadTweets()
                    // on awt thread
                    SwingUtilities.invokeAndWait(this);
                    Thread.sleep(1000*60*3); // 3 mins
                }
                catch( Exception e ) {
                    // break loop
                    break;
                }
            }
        }
    }
    
    private void getUsernameAndPassword() {
        Preferences prefs = Preferences.userNodeForPackage(getClass());
        user = prefs.get(USER_PREF, user);
        password = prefs.get(PASSWORD_PREF, password);
        
        Pref dialog = new Pref(this, user, password);
        
        dialog.pack();
        dialog.setLocationRelativeTo(this);
        dialog.setModal(true);
        dialog.show();
        
        if ( dialog.okPressed ) {
            user = dialog.username.getText();
            password = dialog.password.getText();
            prefs.put(USER_PREF, user);
            prefs.put(PASSWORD_PREF, password);
            try {
                prefs.flush();
            }
            catch(Exception e){}
        }
    }
    
    private Document loadDocument(String urlStr, String status) {
        try {
            statusBar.setText("Connecting");
            
            //String userPassword = user + ":" + password;
            String encoding = new sun.misc.BASE64Encoder().encode((user + ":" + password).getBytes());
            
            URL url = new URL(urlStr);
            URLConnection conn = url.openConnection();
            conn.setRequestProperty ("Authorization", "Basic " + encoding);
            
            if ( status != null ) {
                conn.setDoOutput(true);
                OutputStreamWriter out = new OutputStreamWriter(
                    conn.getOutputStream());
                out.write("status=" + URLEncoder.encode(status,"UTF-8"));
                out.close();
            }
            
            DocumentBuilderFactory f = DocumentBuilderFactory.newInstance();
            DocumentBuilder b = f.newDocumentBuilder();
            
            InputStream in = conn.getInputStream();
            Document doc = b.parse(in);
            in.close();
            
            statusBar.setText(" ");
            
            return doc;
        }
        catch( Exception e ) {
            //e.printStackTrace();
            statusBar.setText(e.getMessage());
        }
        return null;
    }
    
    private void loadTweets() {
        if ( tweetsDoc != null ) {
        
            NodeList list = tweetsDoc.getElementsByTagName("status");
            // run through in reverse order, so new tweets get put into right place
            for ( int i = list.getLength()-1; i >= 0; i-- ) {
                Element status = (Element)list.item(i);
                String id = getNodeText(status.getElementsByTagName("id"));
                if ( tweetIDs.contains(id) )
                    continue;
            
                StringBuilder tweet = new StringBuilder(getNodeText(status.getElementsByTagName("screen_name")));
                tweet.append(": ");
                tweet.append(getNodeText(status.getElementsByTagName("text")));
                JEditorPane label = new JEditorPane(); // editor pane will wrap to fit
                label.setEditable(false);
                label.setText(htmlDecode(tweet.toString()));
                label.setToolTipText(getNodeText(status.getElementsByTagName("created_at")));
                label.setBackground(tweetIDs.size()%2==0?Color.WHITE:new Color(0xFFEEEEEE));
                tweets.add(label,0); // add to top
                tweetIDs.add(id);
            
                // set id of most recent tweet
                lastTweetID=id;
            }
        
            processComponentEvent(null);
        }
    }
    
    protected void processComponentEvent(ComponentEvent e) {
        if ( e != null ) { super.processComponentEvent(e); }
        
        // force width of tweets to match width of window
        for ( int i = 0; i < 3; i++ ) {
            tweets.invalidate();
            tweets.setPreferredSize(null);
            int height = tweets.getPreferredSize().height;
            tweets.setPreferredSize(new Dimension(scrollPane.getViewportBorderBounds().width,height));
            //scrollPane.revalidate();
            scrollPane.validate();
            //validate();
        }
    }
    
    private String htmlDecode(String html) {
        // TODO handle more entities
        return html.replaceAll("&gt;",">").replaceAll("&lt;", "<");
    }
    
    private String getNodeText(NodeList list) {
        return (list.getLength() > 0)? list.item(0).getTextContent() : "";
    }
    
    class Pref extends JDialog implements ActionListener {
        public final JTextField username;
        public final JPasswordField password;
        
        private final JButton ok;
        public boolean okPressed = false;
        
        public Pref(JFrame frame, String user, String password) {
            super(frame, "Preferences");
            this.username = new JTextField(user,10);
            this.password = new JPasswordField(password,10);
        
            JPanel center = new JPanel();
            center.setLayout(new GridLayout(2, 2, 5, 5));
        
            center.add(new JLabel("Username:"));
            center.add(this.username);
            center.add(new JLabel("Password:"));
            center.add(this.password);
        
            add(center, BorderLayout.CENTER);
        
            JPanel bottom = new JPanel();
            ok = new JButton("OK");
            JButton cancel = new JButton("Cancel");
        
            bottom.add(cancel);
            bottom.add(ok);
            
            ok.addActionListener(this);
            cancel.addActionListener(this);
            
            add(bottom, BorderLayout.SOUTH);
        
            getRootPane().setDefaultButton(ok);
        }
        
        public void actionPerformed(ActionEvent e) {
            okPressed = e.getSource() == ok;
            hide();
        }
    
    }
    
    public static void main(String[] args) {
        T client = new T();
        client.pack();
        client.show();
        client.start();
    }
    
}